#! /bin/sh
#
# Copyright by The HDF Group.
# All rights reserved.
#
# This file is part of HDF5.  The full HDF5 copyright notice, including
# terms governing use, modification, and redistribution, is contained in
# the COPYING file, which can be found at the root of the source code
# distribution tree, or in https://www.hdfgroup.org/licenses.
# If you do not have access to either file, you may request a copy from
# help@hdfgroup.org.
#
srcdir=.
TOP_BUILDDIR=../../..

# Determine backward compatibility options enabled
DEPRECATED_SYMBOLS="yes"

EXIT_SUCCESS=0
EXIT_FAILURE=1

H5DIFF=../../src/h5diff/h5diff               # The tool name
H5DIFF_BIN=`pwd`/$H5DIFF    # The path of the tool binary

nerrors=0
verbose=yes
h5haveexitcode=yes        # default is yes

TEST_NAME=ud_plugin
FROM_DIR=`pwd`/.libs
PLUGIN_LIB="$FROM_DIR/libdynlibdiff.*"
PLUGIN_LIBDIR=testdir3
RM='rm -rf'

RM='rm -rf'
CMP='cmp -s'
DIFF='diff -c'
CP='cp'
DIRNAME='dirname'
LS='ls'
AWK='awk'

# source dirs
SRC_TOOLS="$srcdir/../.."

# testfiles source dirs for tools
SRC_H5DIFF_TESTFILES="$SRC_TOOLS/test/h5diff/testfiles"

TESTDIR=./testplug
test -d $TESTDIR || mkdir $TESTDIR

######################################################################
# test files
# --------------------------------------------------------------------
# All the test files copy from source directory to test directory
# NOTE: Keep this framework to add/remove test files.
#       Any test files from other tools can be used in this framework.
#       This list are also used for checking exist.
#       Comment '#' without space can be used.
# --------------------------------------------------------------------
LIST_HDF5_TEST_FILES="
$SRC_H5DIFF_TESTFILES/tudfilter.h5
$SRC_H5DIFF_TESTFILES/tudfilter2.h5
$SRC_H5DIFF_TESTFILES/h5diff_ud.txt
$SRC_H5DIFF_TESTFILES/h5diff_udfail.err
$SRC_H5DIFF_TESTFILES/h5diff_udfail.txt
"

# Main Body
# Create test directories if not exists yet.
test -d $PLUGIN_LIBDIR || mkdir -p $PLUGIN_LIBDIR
if [ $? != 0 ]; then
    echo "Failed to create test directory($PLUGIN_LIBDIR)"
    exit $EXIT_FAILURE
fi

# copy plugin library for test
$CP $PLUGIN_LIB $PLUGIN_LIBDIR
if [ $? != 0 ]; then
    echo "Failed to copy plugin library ($PLUGIN_LIB) for test."
    exit $EXIT_FAILURE
fi

# setup plugin path
ENVCMD="env HDF5_PLUGIN_PATH=../${PLUGIN_LIBDIR}:${HDF5_PLUGIN_PATH}"

#
# copy test files and expected output files from source dirs to test dir
#
COPY_TESTFILES="$LIST_HDF5_TEST_FILES"

COPY_TESTFILES_TO_TESTDIR()
{
    # copy test files. Used -f to make sure get a new copy
    for tstfile in $COPY_TESTFILES
    do
        # ignore '#' comment
        echo $tstfile | tr -d ' ' | grep '^#' > /dev/null
        RET=$?
        if [ $RET -eq 1 ]; then
            # skip cp if srcdir is same as destdir
            # this occurs when build/test performed in source dir and
            # make cp fail
            SDIR=`$DIRNAME $tstfile`
            INODE_SDIR=`$LS -i -d $SDIR | $AWK -F' ' '{print $1}'`
            INODE_DDIR=`$LS -i -d $TESTDIR | $AWK -F' ' '{print $1}'`
            if [ "$INODE_SDIR" != "$INODE_DDIR" ]; then
                $CP -f $tstfile $TESTDIR
                if [ $? -ne 0 ]; then
                    echo "Error: FAILED to copy $tstfile ."

                    # Comment out this to CREATE expected file
                    exit $EXIT_FAILURE
                fi
            fi
        fi
    done
}

CLEAN_TESTFILES_AND_TESTDIR()
{
    # skip rm if srcdir is same as destdir
    # this occurs when build/test performed in source dir and
    # make cp fail
    SDIR=$SRC_H5DIFF_TESTFILES
    INODE_SDIR=`$LS -i -d $SDIR | $AWK -F' ' '{print $1}'`
    INODE_DDIR=`$LS -i -d $TESTDIR | $AWK -F' ' '{print $1}'`
    if [ "$INODE_SDIR" != "$INODE_DDIR" ]; then
        $RM $TESTDIR
    fi
}

# Parse option
#   -p   run ph5diff tests
#   -h   print help page
while [ $# -gt 0 ]; do
    case "$1" in
    -p) # reset the tool name and bin to run ph5diff tests
        TESTNAME=ph5diff
        H5DIFF=../../src/h5diff/ph5diff  # The tool name
        H5DIFF_BIN=`pwd`/$H5DIFF    # The path of the tool binary
        pmode=yes
        shift
        ;;
    -h) # print help page
        echo "$0 [-p] [-h]"
        echo "    -p   run ph5diff tests"
        echo "    -h   print help page"
        shift
        exit 0
        ;;
    *)  # unknown option
        echo "$0: Unknown option ($1)"
        exit 1
        ;;
    esac
done

# RUNSERIAL is used. Check if it can return exit code from executalbe correctly.
if [ -n "$RUNSERIAL_NOEXITCODE" ]; then
    echo "***Warning*** Serial Exit Code is not passed back to shell correctly."
    echo "***Warning*** Exit code checking is skipped."
    h5haveexitcode=no
fi


# Print a $* message left justified in a field of 70 characters
#
MESSAGE() {
   SPACES="                                                               "
   echo "$* $SPACES" | cut -c1-70 | tr -d '\012'
}

# Print a line-line message left justified in a field of 70 characters
# beginning with the word "Testing".
#
TESTING() {
    SPACES="                                                               "
    echo "Testing $* $SPACES" | cut -c1-70 | tr -d '\012'
}

# Print a line-line message left justified in a field of 70 characters
# beginning with the word "Verifying".
#
VERIFY() {
    MESSAGE "Verifying $*"
}

# Source in the output filter function definitions.
. $srcdir/../../../bin/output_filter.sh

# Run a test and print PASS or *FAIL*.  If a test fails then increment
# the `nerrors' global variable and (if $verbose is set) display the
# difference between the actual output and the expected output. The
# expected output is given as the first argument to this function and
# the actual output file is calculated by replacing the `.ddl' with
# `.out'.  The actual output is not removed if $HDF5_NOCLEANUP has a
# non-zero value.
#
# Need eval before the RUNCMD command because some machines like
# AIX, has RUNPARALLEL in the style as
#   MP_PROCS=3 MP_TASKS_PER_NODE=3 poe ./a.out
# that throws the shell script off.
#
TOOLTEST() {
    expect="$TESTDIR/$1"
    actual="$TESTDIR/`basename $1 .txt`.out"
    actual_err="$TESTDIR/`basename $1 .txt`.out.err"
    actual_sav=${actual}-sav
    actual_err_sav=${actual_err}-sav
    shift
    if test -n "$pmode"; then
        RUNCMD=$RUNPARALLEL
    else
        RUNCMD=$RUNSERIAL
    fi

    # Run test.
    TESTING $H5DIFF $@
    (
        #echo "#############################"
        #echo "Expected output for '$H5DIFF $@'"
        #echo "#############################"
        cd $TESTDIR
        eval $ENVCMD $RUNCMD $H5DIFF_BIN "$@"
    ) >$actual 2>$actual_err
    EXIT_CODE=$?

    # save actual and actual_err in case they are needed later.
    cp $actual $actual_sav
    STDOUT_FILTER $actual
    cp $actual_err $actual_err_sav
    STDERR_FILTER $actual_err
    cat $actual_err >> $actual

    # don't add exit code check in pmode, as it causes failure. (exit code
    # is from mpirun not tool)
    # if any problem occurs relate to an exit code, it will be caught in
    # serial mode, so the test is fulfilled.
    if test $h5haveexitcode = 'yes' -a -z "$pmode"; then
      echo "EXIT CODE: $EXIT_CODE" >> $actual
    fi

    if [ ! -f $expect ]; then
        # Create the expect file if it doesn't yet exist.
        echo " CREATED"
        cp $actual $expect
        echo "    Expected result (*.ddl) missing"
        nerrors="`expr $nerrors + 1`"
    elif $CMP $expect $actual; then
        echo " PASSED"
    elif test $h5haveexitcode = 'yes' -a -z "$pmode"; then
        echo "*FAILED*"
        echo "    Expected result ($expect) differs from actual result ($actual)"
        nerrors="`expr $nerrors + 1`"
        test yes = "$verbose" && $DIFF $expect $actual |sed 's/^/    /'
    else
        # parallel mode output are often of different ordering from serial
        # output.  If the sorted expected and actual files compare the same,
        # it is safe to assume the actual output match the expected file.
        expect_sorted=expect_sorted
        actual_sorted=actual_sorted
        sort $expect -o $expect_sorted
        sort $actual -o $actual_sorted

        # remove "EXIT CODE:" line from expect file. test for exit code
        # is done by serial mode.
        grep -v "EXIT CODE:" $expect_sorted > $expect_sorted.noexit
        mv $expect_sorted.noexit $expect_sorted

        if $CMP $expect_sorted $actual_sorted; then
            echo " PASSED"
        else
            echo "*FAILED*"
            nerrors="`expr $nerrors + 1`"
            if test yes = "$verbose"; then
                echo "====Expected result ($expect_sorted) differs from actual result ($actual_sorted)"
                $DIFF $expect_sorted $actual_sorted |sed 's/^/    /'
                echo "====The actual output ($actual_sav)"
                sed 's/^/    /' < $actual_sav
                echo "====The actual stderr ($actual_err_sav)"
                sed 's/^/    /' < $actual_err_sav
                echo "====End of actual stderr ($actual_err_sav)"
                echo ""
            fi
        fi
    fi

    # Clean up output file
    if test -z "$HDF5_NOCLEANUP"; then
      rm -f $actual $actual_err $actual_sav $actual_err_sav
      rm -f $actual_sorted $expect_sorted
    fi
}
# Same as TOOLSET except only err file checked
TOOLTEST_ERR() {
    expect="$TESTDIR/$1"
    expect_err="$TESTDIR/`basename $1 .txt`.err"
    actual="$TESTDIR/`basename $1 .txt`.out"
    actual_err="$TESTDIR/`basename $1 .txt`.out.err"
    actual_sav=${actual}-sav
    actual_err_sav=${actual_err}-sav
    shift
    if test -n "$pmode"; then
        RUNCMD=$RUNPARALLEL
    else
        RUNCMD=$RUNSERIAL
    fi

    # Run test.
    TESTING $H5DIFF $@
    (
        #echo "#############################"
        #echo "Expected output for '$H5DIFF $@'"
        #echo "#############################"
        cd $TESTDIR
        eval $ENVCMD $RUNCMD $H5DIFF_BIN "$@"
    ) >$actual 2>$actual_err
    EXIT_CODE=$?

    # save actual and actual_err in case they are needed later.
    cp $actual $actual_sav
    STDOUT_FILTER $actual
    cp $actual_err $actual_err_sav
    STDERR_FILTER $actual_err

    # don't add exit code check in pmode, as it causes failure. (exit code
    # is from mpirun not tool)
    # if any problem occurs relate to an exit code, it will be caught in
    # serial mode, so the test is fulfilled.
    if test $h5haveexitcode = 'yes' -a -z "$pmode"; then
      echo "EXIT CODE: $EXIT_CODE" >> $actual
    fi

    if [ ! -f $expect ]; then
        # Create the expect file if it doesn't yet exist.
        echo " CREATED"
        cp $actual $expect
        echo "    Expected result (*.ddl) missing"
        nerrors="`expr $nerrors + 1`"
    elif $CMP $expect_err $actual_err; then
        echo " PASSED"
    elif test $h5haveexitcode = 'yes' -a -z "$pmode"; then
        echo "*FAILED*"
        echo "    Expected result ($expect_err) differs from actual result ($actual_err)"
        nerrors="`expr $nerrors + 1`"
        test yes = "$verbose" && $DIFF $expect_err $actual_err |sed 's/^/    /'
    else
        # parallel mode output are often of different ordering from serial
        # output.  If the sorted expect_err and actual_err files compare the same,
        # it is safe to assume the actual output match the expected file.
        expect_sorted=expect_sorted
        actual_sorted=actual_sorted
        sort $expect_err -o $expect_sorted
        sort $actual_err -o $actual_sorted
        mv $expect_sorted.noexit $expect_sorted

        if $CMP $expect_sorted $actual_sorted; then
            echo " PASSED"
        else
            echo "*FAILED*"
            nerrors="`expr $nerrors + 1`"
            if test yes = "$verbose"; then
                echo "====Expected result ($expect_sorted) differs from actual result ($actual_sorted)"
                $DIFF $expect_sorted $actual_sorted |sed 's/^/    /'
                echo "====The actual output ($actual_sav)"
                sed 's/^/    /' < $actual_sav
                echo "====The actual stderr ($actual_err_sav)"
                sed 's/^/    /' < $actual_err_sav
                echo "====End of actual stderr ($actual_err_sav)"
                echo ""
            fi
        fi
    fi

    # Clean up output file
    if test -z "$HDF5_NOCLEANUP"; then
      rm -f $actual $actual_err $actual_sav $actual_err_sav
      rm -f $actual_sorted $expect_sorted
    fi
}

##############################################################################
###           T H E   T E S T S
##############################################################################
# prepare for test
COPY_TESTFILES_TO_TESTDIR

# Run the test
TOOLTEST h5diff_ud.txt -v tudfilter.h5 tudfilter2.h5

# print results
if test $nerrors -ne 0 ; then
    echo "$nerrors errors encountered"
    exit_code=$EXIT_FAILURE
else
    echo "All Plugin API tests passed."
    exit_code=$EXIT_SUCCESS
fi

# Clean up temporary files/directories
CLEAN_TESTFILES_AND_TESTDIR

# Clean up temporary files/directories and leave
$RM $PLUGIN_LIBDIR

exit $exit_code
